# ==================================== #

WELCOME_MESSAGE_TEXT = """
ℹ️ Benvenuto allo spazio Matrix dell'Unimore!

Sono {username_text}, il bot che verifica le credenziali degli utenti che entrano e permette loro di accedere alle rispettive aree.

Se sei uno studente, puoi ottenere accesso all'Area Studenti verificando la tua identità:
{profile_url}
"""

# language=html
WELCOME_MESSAGE_HTML = """
<p>
    ℹ️ Benvenuto allo spazio Matrix dell'Unimore!
</p>
<p>
    Sono {username_html}, il bot buttafuori che verifica le credenziali degli utenti che entrano e permette loro di accedere alle rispettive aree.
</p>
<p>
    Se sei uno studente, puoi ottenere accesso all'Area Studenti <a href="{profile_url}">verificando la tua identità</a>!
</p>
"""

# ==================================== #

SUCCESS_MESSAGE_TEXT = """
✅ Ti sei unito all'Area Studenti: benvenuto!

Se in qualsiasi momento vuoi modificare o eliminare i tuoi dati salvati sul mio database, puoi accedervi dal tuo profilo privato:
{profile_url}
"""

# language=html
SUCCESS_MESSAGE_HTML = """
<p>
    ✅ Ti sei unito all'Area Studenti: benvenuto!
</p>
<p>
    Se in qualsiasi momento vuoi modificare o eliminare i tuoi dati salvati sul mio database, puoi <a href="{profile_url}">accedervi dal tuo profilo privato</a>!
</p>
"""

# ==================================== #

GOODBYE_MESSAGE_TEXT = """
❎ Hai abbandonato lo spazio Matrix dell'Unimore, quindi ho cancellato i tuoi dati dal mio database; a breve verrai inoltre rimosso da tutte le stanze dello spazio.

Se cambierai idea in futuro, potrai sempre rientrare allo stesso indirizzo!

Abbi un buon proseguimento di giornata!
"""

# language=html
GOODBYE_MESSAGE_HTML = """
<p>
    ❎ Hai abbandonato lo spazio Matrix dell'Unimore, quindi ho cancellato i tuoi dati dal mio database; a breve verrai inoltre rimosso da tutte le stanze dello spazio.
</p>
<p>
    Se cambierai idea in futuro, potrai sempre rientrare allo stesso indirizzo!
</p>
<p>
    Abbi un buon proseguimento di giornata!
</p>
"""

# ==================================== #

UNLINK_MESSAGE_TEXT = """
📴 Hai abbandonato l'Area Studenti dello spazio Unimore, quindi ho scollegato il tuo account Studenti@Unimore dal tuo account Matrix.

Se cambierai idea in futuro, potrai sempre essere riaggiunto all'Area Studenti ricollegando il tuo account:
{profile_url}

Abbi un buon proseguimento di giornata!
"""

# language=html
UNLINK_MESSAGE_HTML = """
<p>
    📴 Hai abbandonato l'Area Studenti dello spazio Unimore, quindi ho scollegato il tuo account Studenti@Unimore dal tuo account Matrix.
</p>
<p>
    Se cambierai idea in futuro, potrai sempre essere riaggiunto all'Area Studenti <a href="{profile_url}">ricollegando il tuo account</a>!
</p>
<p>
    Abbi un buon proseguimento di giornata!
</p>
"""

# ==================================== #

__all__ = (
    "WELCOME_MESSAGE_TEXT",
    "WELCOME_MESSAGE_HTML",
    "SUCCESS_MESSAGE_TEXT",
    "SUCCESS_MESSAGE_HTML",
    "GOODBYE_MESSAGE_TEXT",
    "GOODBYE_MESSAGE_HTML",
    "UNLINK_MESSAGE_TEXT",
    "UNLINK_MESSAGE_HTML",
)



import socket


def generate_device_name(name: str) -> str:
    """
    Generate a human-readable device name from the machine's hostname and a given name to use as `display name for Matrix clients <https://spec.matrix.org/latest/#devices>`_.
    """
    return f"Loki [{name}] @ {socket.gethostname()}"


__all__ = (
    "generate_device_name",
)



"""
This module defines some extensions for :mod:`nio`, as it is currently missing some methods that :mod:`lokiunimore` desperately needs.

It also defines a custom client with the functions we need to run :mod:`lokiunimore`.
"""
import contextlib

import aiohttp
import nio
import nio.crypto
import logging
import hashlib
import hmac
import typing as t
import functools
import sqlalchemy
import sqlalchemy.orm

T = t.TypeVar("T")
log = logging.getLogger(__name__)

from lokiunimore.sql.tables import MatrixUser, MatrixProcessedEvent
from lokiunimore.utils.device_names import generate_device_name
from lokiunimore.config import MATRIX_PUBLIC_SPACE_ID, MATRIX_PRIVATE_SPACE_ID, MATRIX_SKIP_EVENTS
from lokiunimore.matrix.templates.messages import WELCOME_MESSAGE_TEXT, WELCOME_MESSAGE_HTML, SUCCESS_MESSAGE_TEXT, SUCCESS_MESSAGE_HTML, GOODBYE_MESSAGE_TEXT, GOODBYE_MESSAGE_HTML, UNLINK_MESSAGE_TEXT, UNLINK_MESSAGE_HTML
from lokiunimore.web.app import app


class RequestError(Exception):
    """
    A request to the Matrix homeserver failed.
    """

    def __init__(self, response: nio.responses.ErrorResponse):
        super().__init__()
        self.response: nio.responses.ErrorResponse = response


# noinspection PyProtectedMember
class ExtendedAsyncClient(nio.AsyncClient):
    """
    An :class:`~nio.AsyncClient` with some extra features to be upstreamed some day.
    """

    def __repr__(self):
        return f"<{self.__class__.__qualname__} for {self.user} at {self.homeserver}>"

    def _send(
            self,
            response_class: t.Type[T],
            method: str,
            path: str,
            data: t.Union[None, str, nio.crypto.AsyncDataT] = None,
            response_data: t.Optional[t.Tuple[t.Any, ...]] = None,
            content_type: t.Optional[str] = None,
            trace_context: t.Optional[t.Any] = None,
            data_provider: t.Optional[nio.DataProvider] = None,
            timeout: t.Optional[float] = None,
            content_length: t.Optional[int] = None,
    ) -> T:
        """
        Override :meth:`nio.client.AsyncClient._send` with something that actually raises errors instead of returning them.

        .. danger::

            SUPER UNTESTED. PLEASE WORK. PLEASE DO. PLEASE DO NOT EXPLODE IN PRODUCTION.
        """
        result = super()._send(
            response_class=response_class,
            method=method,
            path=path,
            data=data,
            response_data=response_data,
            content_type=content_type,
            trace_context=trace_context,
            data_provider=data_provider,
            timeout=timeout,
            content_length=content_length,
        )

        if isinstance(result, nio.responses.ErrorResponse):
            log.warning(f"{method} {path} errored: {result!r}")
            raise RequestError(result)

        return result

    async def register_with_shared_secret(self, shared_secret: str, username: str, displayname: str, password: str):
        """
        Given the homeserver's shared secret, perform a `shared-secret registration <https://matrix-org.github.io/synapse/latest/admin_api/register_api.html#shared-secret-registration>`_.

        :param shared_secret: The shared secret to register with.
        :param username: The username of the user to register.
        :param displayname: The displayname of the user to register.
        :param password: The password of the user to register.
        :raises aiohttp.ClientResponseError: If a request is not successful.

        .. note:: ``shared_secret`` is not the same shared secret of :meth:`.login_with_shared_secret`.
        """

        log.debug(f"Registering via shared-secret registration: {username}")

        path = nio.Api._build_path(
            ["v1", "register"],
            base_path="/_synapse/admin"
        )

        nonce: aiohttp.ClientResponse = await self.send("GET", path)
        nonce.raise_for_status()
        nonce: dict[str, ...] = await nonce.json()
        nonce: str = nonce["nonce"]

        mac = hmac.new(shared_secret.encode("utf8"), digestmod=hashlib.sha1)
        mac.update(nonce.encode("utf8"))
        mac.update(b"\0")
        mac.update(username.encode("utf8"))
        mac.update(b"\0")
        mac.update(password.encode("utf8"))
        mac.update(b"\0")
        mac.update(b"notadmin")

        registration: aiohttp.ClientResponse = await self.send("POST", path, nio.Api.to_json({
            "nonce": nonce,
            "username": username,
            "displayname": displayname,
            "password": password,
            "admin": False,
            "mac": mac.hexdigest()
        }))
        registration.raise_for_status()

        log.info(f"Registered via shared-secret registration: {username}")

    async def login_with_shared_secret(self, shared_secret: str) -> None:
        """
        Given a shared secret, perform a login via `com.devture.shared_secret_auth <https://github.com/devture/matrix-synapse-shared-secret-auth>`_.

        :param shared_secret: The shared secret to login with.

        .. note:: ``shared_secret`` is not the same shared secret of :meth:`.register_with_shared_secret`.
        """

        log.debug(f"Logging in as {self.user} with a shared secret...")

        token = hmac.new(key=shared_secret.encode("utf8"), msg=self.user.encode("utf8"), digestmod=hashlib.sha512).hexdigest()

        await self.login_raw({
            "type": "com.devture.shared_secret_auth",
            "identifier": {
                "type": "m.id.user",
                "user": self.user,
            },
            "token": token,
            "initial_device_display_name": generate_device_name(__name__),
        })

        log.debug(f"Login successful!")

    async def room_hierarchy(self, room_id: str, max_depth: int, suggested_only: bool) -> list[dict]:
        """
        Given a space, get the room hierarchy.

        :param room_id: The id of the space to get the hierarchy of.
        :param max_depth: The max depth of subspaces to recurse into.
        :param suggested_only: Whether only suggested rooms should be returned.
        :return:
        """

        log.debug(f"Getting room hierarchy for: {room_id!r}")

        current = None
        rooms = []

        while True:
            path = ["rooms", room_id, "hierarchy"]
            query = {
                "max_depth": str(max_depth),
                "suggested_only": str(suggested_only).lower(),
            }
            if current:
                query["from"] = current

            result: aiohttp.ClientResponse = await self.send(
                "GET",
                nio.Api._build_path(path, query, base_path="/_matrix/client/v1"),
                headers={
                    "Authorization": f"Bearer {self.access_token}"
                }
            )
            result: dict = await result.json()

            rooms = [*rooms, *result["rooms"]]

            current = result.get("next_batch")
            if not current:
                break

        log.debug(f"Successfully retrieved a hierarchy of {len(rooms)} rooms!")

        return rooms

    async def put_management_room(self, user_id: str) -> str:
        """
        Find the first available management room with the given user, or create one if none exist.

        :param user_id: The user to message.
        :returns: The room id of the created room.
        """

        log.debug(f"Creating a management room for %s", user_id)

        for room in self.rooms.values():
            is_dm = "m.direct" in room.tags
            is_group = room.is_group
            has_two_users = len(room.users) + len(room.invited_users) == 2
            contains_user_id = user_id in room.users.keys() or user_id in room.invited_users.keys()

            if (is_dm or is_group) and has_two_users and contains_user_id:
                log.debug(f"Found existing management room %s for %s", room.room_id, user_id)
                return room.room_id

        else:
            log.debug(f"Creating new management room for %s", user_id)
            response = await self.room_create(invite=[user_id], is_direct=True)

            if not isinstance(response, nio.RoomCreateResponse):
                raise Exception("Failed to create a management room.")

            log.info(f"Created new managememt room %s for %s", response.room_id, user_id)
            return response.room_id

    async def room_send_message_html(self, room_id: str, text: str, html: str):
        """
        Send an HTML message with a text fallback.

        :param room_id: The ID of the room to send the message in.
        :param text: The plain text fallback of the message.
        :param html: The HTML message.
        :return: I have no idea, whatever :meth:`~nio.client.AsyncClient.room_send` does.
        """

        return await self.room_send(room_id, "m.room.message", {
            "msgtype": "m.notice",
            "format": "org.matrix.custom.html",
            "formatted_body": html,
            "body": text,
        })

    async def mention_html(self, user_id: str) -> str:
        """
        Create a rich HTML mention.

        To provide a text fallback for the HTML mention, use the ``user_id``.

        :param user_id: The user to mention.
        :return: The HTML string.
        """

        display_name = await self.get_displayname(user_id)

        # language=html
        return f"""<a href="https://matrix.to/#/{user_id}">{display_name}</a>"""


def filter_processed_events(f):
    """
    Decorator applicable to a :mod:`nio` callback to filter incoming events whose IDs have been marked in the database as *already processed*, and marking events successfully processed by the function as "processed".
    """

    @functools.wraps(f)
    async def wrapped(self, room: nio.MatrixRoom, event: nio.Event):
        if event_id := getattr(event, "event_id", None):
            log.debug(f"Checking if event should be processed: {event_id}")
            if self._event_processed_check(event):
                if not MATRIX_SKIP_EVENTS.__wrapped__:
                    log.debug(f"Processing event: {event_id}")
                    await f(self, room, event)
                else:
                    log.debug(f"Skipping event due to MATRIX_SKIP_EVENTS: {event_id}")
                log.debug(f"Marking event as processed: {event_id}")
                self._event_processed_mark(event)
            else:
                log.debug(f"Skipping already processed event: {event_id}")
        else:
            log.debug(f"Received partial event with no id, which should always be processed")
            await f(self, room, event)
            log.debug(f"Processed partial event successfully")

    return wrapped


class LokiClient(ExtendedAsyncClient):
    def __init__(self, *args, database_url: str, **kwargs):
        super().__init__(*args, **kwargs)

        self.sqla_engine: sqlalchemy.engine.Engine = sqlalchemy.create_engine(database_url)
        """
        The :mod:`sqlalchemy` :class:`~sqlalchemy.engine.Engine` associated with this client.
        """

        # noinspection PyTypeChecker
        self.add_event_callback(self.__handle_membership_change, nio.InviteMemberEvent)
        # noinspection PyTypeChecker
        self.add_event_callback(self.__handle_membership_change, nio.RoomMemberEvent)

    @contextlib.contextmanager
    def _sqla_session(self) -> t.Generator[sqlalchemy.orm.Session, None, None]:
        """
        Open a new :mod:`sqlalchemy` :class:`~sqlalchemy.orm.Session` using this object's :attr:`.sql_engine` via a context manager interface.
        """

        with sqlalchemy.orm.Session(bind=self.sqla_engine) as session:
            yield session

    def _event_processed_check(self, event: nio.Event) -> bool:
        """
        Check if a received event should be processed.

        :param event: The event to check.
        :return: :data:`True` if the event should be processed, :data:`False` otherwise.
        """

        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            return session.query(MatrixProcessedEvent).get(event.event_id) is None

    def _event_processed_mark(self, event: nio.Event):
        """
        Mark the event as processed, preventing it from being processed again.

        :param event: The event to mark.
        :return: The created and committed :class:`lokiunimore.sql.tables.MatrixProcessedEvent` instance.
        """

        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            mpe: MatrixProcessedEvent = MatrixProcessedEvent(id=event.event_id)
            session.add(mpe)
            session.commit()
            return mpe

    @filter_processed_events
    async def __handle_membership_change(self, room: nio.MatrixRoom, event: nio.Event) -> None:
        # Filters allow us to determine the event type in a better way
        event: nio.InviteMemberEvent | nio.RoomMemberEvent

        if event.prev_membership == event.membership:
            # If the event is a name change, or something like that, don't do anything
            return

        # Catch events about myself immediately
        elif event.state_key == self.user_id:
            # If I'm invited to a room, join it
            if event.membership == "invite":
                await self.__handle_received_invite(room)

        elif event.membership == "join":
            # If somebody joins the parent space, notify them of my presence and send them their profile URL
            if room.room_id == MATRIX_PUBLIC_SPACE_ID:
                await self.__handle_public_space_joiner(event.state_key)
            # If somebody joins the child space, notify them of the successful login
            elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
                await self.__handle_private_space_joiner(event.state_key)

        elif event.membership == "leave":
            # If somebody leaves the parent space, remove them from all subrooms, delete their account, and finally notify them
            if room.room_id == MATRIX_PUBLIC_SPACE_ID:
                await self.__handle_public_space_leaver(event.state_key)
            # If somebody leaves the child space, remove them from all subrooms, delete their linking, and finally notify them
            elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
                await self.__handle_private_space_leaver(event.state_key)

        elif event.membership == "ban":
            # TODO: If somebody is banned from the parent space... propagate the ban?
            if room.room_id == MATRIX_PUBLIC_SPACE_ID:
                await self.__handle_public_space_leaver(event.state_key)
            # TODO: If somebody is banned from the child space... propagate the ban?
            elif room.room_id == MATRIX_PRIVATE_SPACE_ID:
                await self.__handle_private_space_leaver(event.state_key)

    async def __handle_received_invite(self, room: nio.MatrixRoom):
        log.debug(f"Received invite to: {room.room_id}")
        await self.join(room.room_id)
        log.info(f"Accepted invite to: {room.room_id}")

    async def __handle_public_space_joiner(self, user_id: str):
        log.debug(f"User joined public space: {user_id}")

        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            matrix_user: MatrixUser = MatrixUser.create(session=session, id=user_id)
            session.commit()

            with app.app_context():
                formatting = dict(
                    username_text=self.user_id,
                    username_html=await self.mention_html(self.user_id),
                    profile_url=matrix_user.profile_url(),
                )

        log.debug(f"Notifying user of the account creation: {user_id}")
        await self.room_send_message_html(
            await self.put_management_room(user_id),
            text=WELCOME_MESSAGE_TEXT.format(**formatting),
            html=WELCOME_MESSAGE_HTML.format(**formatting)
        )
        log.debug(f"Notified user of the account creation: {user_id}")

        log.info(f"Handled joiner of public space: {user_id}")

    async def __handle_private_space_joiner(self, user_id: str):
        log.info(f"User joined private space: {user_id}")

        log.debug(f"Setting MatrixUser as joined for: {user_id}")
        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
            if matrix_user is None:
                log.warning(f"User joined private space without having a pre-existent record in the db: {user_id}")
                matrix_user = MatrixUser.create(session=session, id=user_id)
                session.commit()

            matrix_user.joined_private_space = True
            session.commit()
            log.debug(f"Set MatrixUser as joined for: {user_id}")

            with app.app_context():
                formatting = dict(
                    profile_url=matrix_user.profile_url(),
                )

        log.debug(f"Notifying user of the account link: {user_id}")
        await self.room_send_message_html(
            await self.put_management_room(user_id),
            text=SUCCESS_MESSAGE_TEXT.format(**formatting),
            html=SUCCESS_MESSAGE_HTML.format(**formatting)
        )
        log.debug(f"Notified user of the account link: {user_id}")

        log.info(f"Handled joiner of private space: {user_id}")

    async def __handle_public_space_leaver(self, user_id: str):
        log.info(f"User left public space: {user_id}")

        log.debug(f"Deleting MatrixUser for: {user_id}")
        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
            if matrix_user is None:
                log.warning(f"User left public space without having a pre-existent record in the db: {user_id}")
            else:
                matrix_user.destroy(session=session)
                session.commit()

        log.debug(f"Notifying user of the account deletion: {user_id}")
        await self.room_send_message_html(
            await self.put_management_room(user_id),
            text=GOODBYE_MESSAGE_TEXT,
            html=GOODBYE_MESSAGE_HTML
        )
        log.debug(f"Notified user of the account deletion: {user_id}")

        log.debug(f"Finding room hierarchy of the public space...")
        public_hierarchy = await self.room_hierarchy(MATRIX_PUBLIC_SPACE_ID.__wrapped__, max_depth=9, suggested_only=False)

        log.debug(f"Finding room hierarchy of the private space...")
        private_hierarchy = await self.room_hierarchy(MATRIX_PRIVATE_SPACE_ID.__wrapped__, max_depth=9, suggested_only=False)

        hierarchy = [*public_hierarchy, *private_hierarchy]

        log.debug(f"Removing public space leaver from {len(hierarchy)} rooms: {user_id}")
        success_count = 0
        for room in hierarchy:
            room_id = room["room_id"]
            try:
                await self.room_kick(room_id=room_id, user_id=user_id, reason="Loki account deleted")
            except RequestError as e:
                log.warning(f"Could not remove public space leaver {user_id} from {room_id}: {e!r}")
            else:
                success_count += 1
        log.debug(f"Removed public space leaver from {success_count} rooms: {user_id}")

        log.info(f"Handled leaver of public space: {user_id}")

    async def __handle_private_space_leaver(self, user_id: str):
        log.debug(f"User left private space: {user_id}")

        log.debug(f"Unlinking account for: {user_id}")
        with self._sqla_session() as session:
            session: sqlalchemy.orm.Session
            matrix_user: MatrixUser = session.query(MatrixUser).get(user_id)
            if matrix_user is None:
                log.warning(f"User left private space without having a pre-existent record in the db: {user_id}")
                return
            elif matrix_user.account is None:
                log.warning(f"User left private space without having a linked account in the db: {user_id}")
            else:
                matrix_user.unlink(session=session)
                session.commit()

            with app.app_context():
                formatting = dict(
                    profile_url=matrix_user.profile_url(),
                )

        log.debug(f"Notifying user of the account unlinking: {user_id}")
        await self.room_send_message_html(
            await self.put_management_room(user_id),
            text=UNLINK_MESSAGE_TEXT.format(**formatting),
            html=UNLINK_MESSAGE_HTML.format(**formatting)
        )
        log.debug(f"Notified user of the account unlinking: {user_id}")

        log.debug(f"Finding room hierarchy of the private space...")
        hierarchy = await self.room_hierarchy(MATRIX_PRIVATE_SPACE_ID.__wrapped__, max_depth=9, suggested_only=False)

        log.debug(f"Removing private space leaver from {len(hierarchy)} rooms: {user_id}")
        success_count = 0
        for room in hierarchy:
            room_id = room["room_id"]
            try:
                await self.room_kick(room_id=room_id, user_id=user_id, reason="Loki account unlinked")
            except RequestError as e:
                log.warning(f"Could not remove private space leaver {user_id} from {room_id}: {e}")
            else:
                success_count += 1
        log.debug(f"Removed private space leaver from {success_count} rooms: {user_id}")

        log.info(f"Handled leaver of private space: {user_id}")



import re
import cfig
import sqlalchemy.engine
import dotenv


dotenv.load_dotenv()
config = cfig.Configuration()


@config.required()
def MATRIX_HOMESERVER(val: str) -> str:
    """
    The Matrix homeserver to connect to.
    """
    return val


@config.required()
def MATRIX_USER_ID(val: str) -> str:
    """
    The full user ID of the Matrix account to login as.
    https://spec.matrix.org/latest/#users
    """
    if not val.startswith("@"):
        raise ValueError("User IDs must start with `@`")
    if ":" not in val:
        raise ValueError("User IDs must contain the `:` localpart-server separator")
    return val


@config.required()
def MATRIX_USER_SECRET(val: str) -> str:
    """
    The shared secret to use for logging in to the server via com.devture.shared_secret_auth.
    https://github.com/devture/matrix-synapse-shared-secret-auth
    """
    return val


@config.optional()
def MATRIX_SKIP_EVENTS(val: str | None) -> bool:
    """
    Set this to `true` to have the Matrix client skip processing all events happened until now.
    Useful for initializing a new database on an existing bot.
    """
    return val and val.lower() == "true"


@config.required()
def MATRIX_PUBLIC_SPACE_ID(val: str) -> str:
    """
    The room ID of the public Matrix space to monitor.
    https://spec.matrix.org/latest/#room-structure
    """
    if not val.startswith("!"):
        raise ValueError("Room IDs must start with `!`")
    if ":" not in val:
        raise ValueError("Room IDs must contain the `:` localpart-server separator")
    return val


@config.required()
def MATRIX_PUBLIC_SPACE_ALIAS(val: str) -> str:
    """
    The room alias of the public Matrix space to display in various parts of the bot.
    https://spec.matrix.org/latest/#room-aliases
    """
    if not val.startswith("#"):
        raise ValueError("Room aliases must start with `#`")
    if ":" not in val:
        raise ValueError("Room aliases must contain the `:` localpart-server separator")
    return val


@config.required()
def MATRIX_PRIVATE_SPACE_ID(val: str) -> str:
    """
    The room ID of the private Matrix space to monitor and add users to.
    https://spec.matrix.org/latest/#room-structure
    """
    if not val.startswith("!"):
        raise ValueError("Room IDs must start with `!`")
    if ":" not in val:
        raise ValueError("Room IDs must contain the `:` localpart-server separator")
    return val


@config.required()
def MATRIX_HELP_ROOM_ALIAS(val: str) -> str:
    """
    The room alias of the public Matrix room to request assistance in, to display in various parts of the bot.
    https://spec.matrix.org/latest/#room-aliases
    """
    if not val.startswith("#"):
        raise ValueError("Room aliases must start with `#`")
    if ":" not in val:
        raise ValueError("Room aliases must contain the `:` localpart-server separator")
    return val


@config.required()
def SQLALCHEMY_DATABASE_URL(val: str) -> sqlalchemy.engine.URL:
    """
    The URL of the database to store data in.
    https://docs.sqlalchemy.org/en/14/core/engines.html#database-urls
    """
    return sqlalchemy.engine.url.make_url(val)


@config.required()
def FLASK_SECRET_KEY(val: str) -> str:
    """
    The secret key to use with Flask to sign session cookies.
    https://flask.palletsprojects.com/en/2.2.x/config/#SECRET_KEY
    """
    return val


@config.required()
def FLASK_SERVER_NAME(val: str) -> str:
    """
    The hostname of the server where the Flask webserver is available at.
    For example, `loki.steffo.eu`.
    """
    return val


@config.required()
def FLASK_APPLICATION_ROOT(val: str) -> str:
    """
    The path where the root of the Flask webserver will be served at.
    For example, `/`.
    """
    return val


@config.required()
def FLASK_PREFERRED_URL_SCHEME(val: str) -> str:
    """
    The protocol to prefer when building URLs outside a Flask request.
    For example, `https`
    """
    return val


@config.required()
def OIDC_CLIENT_ID(val: str) -> str:
    """
    The client ID of the OpenID Connect client used to authenticate user sign-ins.
    """
    return val


@config.required()
def OIDC_CLIENT_SECRET(val: str) -> str:
    """
    The client secret of the OpenID Connect client used to authenticate user sign-ins.
    """
    return val


@config.required()
def OIDC_CONFIGURATION_URL(val: str) -> str:
    """
    The URL to the configuration of the OpenID Connect client used to authenticate user sign-ins.
    For example, Google uses `https://accounts.google.com/.well-known/openid-configuration`.
    """
    return val


@config.required()
def OIDC_API_BASE_URL(val: str) -> str:
    """
    The base URL of the API that user sign-ins via the OpenID Connect client grant access to.
    For example, Google uses `https://www.googleapis.com/`.
    """
    return val


@config.optional()
def OIDC_SCOPES(val: str | None) -> str:
    """
    The space-separated additional scopes that the OpenID Connect client should request access to.
    The `openid` scope is added automatically.
    Defaults to requesting the `email profile` scopes.
    """
    if not val:
        return "openid email profile"
    else:
        return "openid " + val


@config.required()
def OIDC_EMAIL_REGEX(val: str) -> re.Pattern:
    """
    A regular expression to match the emails of the users who sign-in via the OpenID Connect client.
    If the email does not match the pattern, the sign-in is rejected.
    For example, `.+@studenti[.]unimore[.]it` to verify that an user is a Unimore student.
    To allow any email, you can use the `.+` regex.
    """
    return re.compile(val)


__all__ = (
    "config",
    "MATRIX_HOMESERVER",
    "MATRIX_USER_ID",
    "MATRIX_USER_SECRET",
    "MATRIX_SKIP_EVENTS",
    "MATRIX_PUBLIC_SPACE_ID",
    "MATRIX_PUBLIC_SPACE_ALIAS",
    "MATRIX_PRIVATE_SPACE_ID",
    "MATRIX_HELP_ROOM_ALIAS",
    "SQLALCHEMY_DATABASE_URL",
    "FLASK_SECRET_KEY",
    "FLASK_SERVER_NAME",
    "FLASK_APPLICATION_ROOT",
    "FLASK_PREFERRED_URL_SCHEME",
    "OIDC_CLIENT_ID",
    "OIDC_CLIENT_SECRET",
    "OIDC_CONFIGURATION_URL",
    "OIDC_API_BASE_URL",
    "OIDC_SCOPES",
    "OIDC_EMAIL_REGEX",
)



import asyncio

from lokiunimore.utils.logs import install_log_handler
from lokiunimore.matrix.client import LokiClient
from lokiunimore.config import MATRIX_USER_SECRET, MATRIX_HOMESERVER, MATRIX_USER_ID, SQLALCHEMY_DATABASE_URL

install_log_handler()
loop = asyncio.new_event_loop()


client = LokiClient(
    homeserver=MATRIX_HOMESERVER.__wrapped__,
    user=MATRIX_USER_ID.__wrapped__,
    database_url=SQLALCHEMY_DATABASE_URL.__wrapped__
)


async def main():
    await client.login_with_shared_secret(MATRIX_USER_SECRET.__wrapped__)
    await client.sync_forever(60_000, full_state=True, set_presence="online")


async def cleanup():
    await client.logout()


try:
    loop.run_until_complete(main())
finally:
    loop.run_until_complete(cleanup())



import logging
import coloredlogs

log = logging.getLogger(__name__)


def install_log_handler(logger: logging.Logger = None):
    if logger is None:
        logger = logging.getLogger("lokiunimore")

    coloredlogs.install(
        logger=logger,
        level="DEBUG",
        fmt="{asctime} | {name:<32} | {levelname:>8} | {message}",
        style="{",
        level_styles=dict(
            debug=dict(color="white"),
            info=dict(color="cyan"),
            warning=dict(color="yellow"),
            error=dict(color="red"),
            critical=dict(color="red", bold=True),
        ),
        field_styles=dict(
            asctime=dict(color='magenta'),
            levelname=dict(color='blue', bold=True),
            name=dict(color='blue'),
        ),
        isatty=True,
    )
    log.info("Installed custom log handler!")


__all__ = (
    "install_log_handler",
)


from lokiunimore.sql.tables import Base
import sqlalchemy
from lokiunimore.config import SQLALCHEMY_DATABASE_URL


sqla_engine: sqlalchemy.engine.Engine = sqlalchemy.create_engine(SQLALCHEMY_DATABASE_URL.__wrapped__)
Base.metadata.create_all(bind=sqla_engine)



import re
import flask
import flask_sqlalchemy
import werkzeug.middleware.proxy_fix
import werkzeug.exceptions
import authlib.integrations.flask_client
import authlib.integrations.base_client
import requests

from lokiunimore.config import config, FLASK_SECRET_KEY, SQLALCHEMY_DATABASE_URL, FLASK_SERVER_NAME, FLASK_APPLICATION_ROOT, FLASK_PREFERRED_URL_SCHEME
from lokiunimore.sql.tables import Base as TableDeclarativeBase
from lokiunimore.sql.tables import Account, MatrixUser
from lokiunimore.web.extensions.matrix_client import MatrixClientExtension


app = flask.Flask(__name__)
"""
The main :mod:`flask` application object.
"""

rp_app = werkzeug.middleware.proxy_fix.ProxyFix(app=app, x_for=1, x_proto=1, x_host=1, x_port=0, x_prefix=0)
"""
Reverse proxied instance of :data:`.app`, to use in production with a Caddy server.
"""

app.config.update({
    **config.proxies.resolve(),
    "SERVER_NAME": FLASK_SERVER_NAME.__wrapped__,
    "APPLICATION_ROOT": FLASK_APPLICATION_ROOT.__wrapped__,
    "PREFERRED_URL_SCHEME": FLASK_PREFERRED_URL_SCHEME.__wrapped__,
    "SECRET_KEY": FLASK_SECRET_KEY.__wrapped__,
    "SQLALCHEMY_DATABASE_URI": SQLALCHEMY_DATABASE_URL.__wrapped__,
    "SQLALCHEMY_TRACK_MODIFICATIONS": False,
})


sqla_extension = flask_sqlalchemy.SQLAlchemy(app=app, metadata=TableDeclarativeBase.metadata)
"""
:mod:`sqlalchemy` database engine, usable by the whole :data:`.app`.
"""


oauth_extension = authlib.integrations.flask_client.OAuth(app=app)
"""
OAuth2 :mod:`flask` extension installed on :data:`.app`.
"""

oauth_extension.register(
    name="oidc",
    server_metadata_url=app.config["OIDC_CONFIGURATION_URL"],
    api_base_url=app.config["OIDC_API_BASE_URL"],
    client_kwargs={
        "scope": app.config["OIDC_SCOPES"]
    },
)


matrix_extension = MatrixClientExtension(app=app)
"""
Synchronous Matrix client for use with Flask.
"""


### Setup the app routes

@app.route("/")
def page_root():
    return flask.render_template("root.html")


@app.route("/privacy")
def page_privacy():
    return flask.render_template("privacy.html")


@app.route("/matrix/<token>/")
def page_matrix_profile(token):
    user: MatrixUser = sqla_extension.session.query(MatrixUser).filter_by(token=token).first_or_404()
    if user.account is None:
        return flask.render_template("matrix/verify.html", user=user, token=token)
    elif not user.joined_private_space:
        return flask.render_template("matrix/join.html", user=user, token=token)
    else:
        return flask.render_template("matrix/complete.html", user=user, token=token)


@app.route("/matrix/<token>/link")
def page_matrix_link(token):
    sqla_extension.session.query(MatrixUser).filter_by(token=token).first_or_404()
    flask.session["matrix_token"] = token
    return oauth_extension.oidc.authorize_redirect(flask.url_for("page_oidc_authorize", _external=True))


@app.route("/matrix/<token>/invite")
def page_matrix_invite(token):
    matrix_user: MatrixUser = sqla_extension.session.query(MatrixUser).filter_by(token=token).first_or_404()

    app.logger.debug(f"Sending private space invite to: {matrix_user.id}")
    try:
        matrix_extension.room_invite(room_id=app.config["MATRIX_PRIVATE_SPACE_ID"], user_id=matrix_user.id)
    except requests.exceptions.HTTPError as e:
        app.logger.warning(f"Failed to send private space invite to {matrix_user.id}: {e}")
        if e.response.status_code == 403:
            return flask.render_template("errors/failed-invite.html"), 500
        if e.response.status_code == 429:
            return flask.render_template("errors/rate-invite.html"), 500

    app.logger.info(f"Sent private space invite to: {matrix_user.id}")

    return flask.redirect(flask.url_for("page_matrix_profile", token=token))


@app.route("/authorize")
def page_oidc_authorize():
    try:
        token = oauth_extension.oidc.authorize_access_token()
    except (authlib.integrations.base_client.errors.OAuthError, werkzeug.exceptions.BadRequestKeyError):
        return flask.render_template("errors/oidc.html"), 400

    # Not sure of why the nonce is now required or if I'm using it correctly
    account = oauth_extension.oidc.parse_id_token(token=token, nonce=token["userinfo"]["nonce"])

    if not account.email_verified:
        return flask.render_template("errors/not-verified.html"), 403

    # noinspection PyTypeChecker
    if not re.match(app.config["OIDC_EMAIL_REGEX"], account.email):
        return flask.render_template("errors/not-student.html"), 403

    local_account = sqla_extension.session.merge(Account(
        email=account.email,
        first_name=account.given_name,
        last_name=account.family_name,
    ))

    if matrix_token := flask.session.pop("matrix_token", None):
        matrix_user = sqla_extension.session.query(MatrixUser).filter_by(token=matrix_token).first_or_404()
        matrix_user.link(session=sqla_extension.session, account=local_account)
        sqla_extension.session.commit()

        return flask.redirect(flask.url_for("page_matrix_invite", token=matrix_token))

    else:
        return flask.render_template("errors/no-link.html"), 403


@app.errorhandler(404)
def page_404(_):
    return flask.render_template("errors/not-found.html"), 404


__all__ = (
    "app",
    "rp_app",
    "sqla_extension",
    "oauth_extension",
    "oauth_extension",
    "matrix_extension",
    "page_root",
    "page_privacy",
    "page_matrix_profile",
    "page_matrix_link",
    "page_matrix_invite",
    "page_oidc_authorize",
)



import flask.logging
import lokiunimore.utils.logs

from .app import app

lokiunimore.utils.logs.install_log_handler(app.logger)
app.logger.removeHandler(flask.logging.default_handler)

app.run(host="127.0.0.1", port=30008, debug=True)



import sqlalchemy as s
import sqlalchemy.orm as o
import secrets
import logging
import flask

log = logging.getLogger(__name__)


Base = o.declarative_base()
"""
The declarative base of all the SQL tables.
"""


class Account(Base):
    """
    A logged-in OpenID Connect account.
    """

    __tablename__ = "accounts"

    email = s.Column(s.String, primary_key=True)
    """
    The verified email of the account.
    """

    first_name = s.Column(s.String)
    """
    The first name of the account owner.
    """

    last_name = s.Column(s.String)
    """
    The last name of the account owner.
    """

    matrix_users = o.relationship("MatrixUser", back_populates="account")
    """
    The Matrix users linked to this account.
    """

    def __repr__(self):
        return f"{self.__qualname__}(id={self.id!r}, email={self.email!r}, first_name={self.first_name!r}, last_name={self.last_name!r}, private={self.private!r})"


class MatrixUser(Base):
    """
    A Matrix user, which may or may not be linked to an account.

    Has an associated token, which allows it to authenticate with Loki.
    """

    __tablename__ = "matrix_users"

    id = s.Column(s.String, primary_key=True)
    """
    The Matrix id of the user, such as ``@steffo:ryg.one``.
    """

    token = s.Column(s.String, nullable=False, default=secrets.token_urlsafe)
    """
    A secure token that the user can use to access their account.
    """

    account_email = s.Column(s.String, s.ForeignKey("accounts.email"))
    """
    If the user linked a OpenID Connect account, its email.
    """

    joined_private_space = s.Column(s.Boolean, nullable=False, default=False)
    """
    Whether this specific Matrix user has joined the private Matrix space monitored by Loki.
    """

    account = o.relationship("Account", back_populates="matrix_users")
    """
    The account linked with this Matrix user.
    """

    def __repr__(self):
        return f"{self.__class__.__qualname__}(id={self.id!r}, token={self.token!r}, account_email={self.account_email!r})"

    @classmethod
    def create(cls, session: o.Session, id: str) -> "MatrixUser":
        """
        Create a `.MatrixUser` in the database for the given Matrix user ID.

        :param session: The `sqlalchemy.orm.Session` to use.
        :param id: The user ID to create an account for.
        :return: The created `.MatrixUser`.
        """

        log.debug("Creating MatrixUser for %s", id)
        matrix_user = MatrixUser(id=id)
        matrix_user = session.merge(matrix_user)
        log.debug("Created MatrixUser for %s", id)
        return matrix_user

    def destroy(self, session: o.Session) -> None:
        """
        Destroy a `.MatrixUser` in the database.

        :param session: The `sqlalchemy.orm.Session` to use.
        """

        id = self.id
        log.debug("Deleting MatrixUser for %s", id)
        if self.account is not None and len(self.account.matrix_users) == 1:
            session.delete(self.account)
        session.delete(self)
        log.debug("Deleted MatrixUser for %s", id)

    # noinspection PyUnusedLocal
    def link(self, session: o.Session, account: Account) -> None:
        """
        Link a `.MatrixUser` to an `.Account`.

        :param session: The `sqlalchemy.orm.Session` to use. (Currently unused.)
        :param account: The `.Account` to link to.
        """
        log.debug("Linking MatrixUser %s to Account %s", self.id, account.email)
        self.account = account
        log.debug("Linked MatrixUser %s to Account %s", self.id, account.email)

    def unlink(self, session: o.Session) -> None:
        """
        Unlink a `.MatrixUser` from an `.Account`, deleting the `.Account` if it has no other connected `.MatrixUser`\\ s.

        :param session: The `sqlalchemy.orm.Session` to use.
        """

        log.debug("Unlinking Account for %s", self.id)
        if len(self.account.matrix_users) == 1:
            session.delete(self.account)
        self.account = None
        self.joined_private_space = False
        log.debug("Unlinked Account for %s", self.id)

    def profile_url(self) -> str:
        """
        Uses the `.token` to get the `~flask.url_for` this `.MatrixUser`'s profile.

        .. warning::
            Must be called inside a `~flask.Flask.app_context`!

        :return: The URL to this `.MatrixUser`'s profile.
        """

        return flask.url_for("page_matrix_profile", token=self.token)


class MatrixProcessedEvent(Base):
    """
    A Matrix event that has been processed by the bot and that should not be processed again.
    """

    __tablename__ = "matrix_processed_events"

    id = s.Column(s.String, primary_key=True)
    """
    The id of the processed event, such as ``$IhC83CM3TRkPG7UbNRsH_t_O2J5DASqzkUYVkPxYR-o``.
    """


__all__ = (
    "Base",
    "Account",
    "MatrixUser",
    "MatrixProcessedEvent",
)
